use std::path::PathBuf;
use crypto::sha1::Sha1;
use crypto::digest::Digest;
use crate::utils::path;
use crate::store::object::{blob::Blob, tree::Tree};
use crate::store::head;
use crate::commands::status::{State, LocalObj};


enum ObjType {
    TREE,
    BLOB,
    DEFAULT
}

pub trait ObjMethods {
    fn get_type(&self) -> ObjType;
    fn get_obj_path(&self) -> PathBuf;
    fn get_file_path(&self) -> PathBuf;
    fn get_relative_file_path(&self) -> PathBuf;
    fn get_name(&self) -> String;
    fn get_hash_path(&self) -> String;
    fn get_local_obj(&self) -> LocalObj;
}

pub struct Obj {
    name: String,
    ref_path: Option<PathBuf>,
    obj_path: PathBuf,
    obj_type: ObjType,
    file_path: PathBuf, // file here is used as both file and directory
    relative_file_path: PathBuf,
    hash_path: String, // hash of the relative path of the file
    //tree: Option<&Tree>,
    //blob: Option<&Blob>
}


impl ObjMethods for Obj {
    fn get_type(&self) -> ObjType {
        self.obj_type
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj_path
    }

    fn get_file_path(&self) -> PathBuf {
        self.file_path
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.relative_file_path
    }

    fn get_local_obj(&self) -> LocalObj {
        LocalObj {
            otype: match self.obj_type {
                ObjType::BLOB => String::from("blob"),
                ObjType::TREE => String::from("tree"),
                ObjType::DEFAULT => String::from("default"),
            },
            name: self.name,
            path: self.file_path,
            path_from: None,
            state: State::New
        }
    }

    fn get_name(&self) -> String {
        self.name
    }
    fn get_hash_path(&self) -> String {
        self.hash_path
    }
}

impl ObjMethods for Blob {
    fn get_type(&self) -> ObjType {
        self.obj.get_type()
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj.get_obj_path()
    }

    fn get_file_path(&self) -> PathBuf {
        self.obj.get_file_path()
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.obj.get_relative_file_path()
    }

    fn get_local_obj(&self) -> LocalObj {
        self.obj.get_local_obj()
    }

    fn get_name(&self) -> String {
        self.obj.get_name()
    }

    fn get_hash_path(&self) -> String {
        self.obj.get_hash_path()
    }
}

impl ObjMethods for Tree {
    fn get_type(&self) -> ObjType {
        self.obj.get_type()
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj.get_obj_path()
    }

    fn get_file_path(&self) -> PathBuf {
        self.obj.get_file_path()
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.obj.get_relative_file_path()
    }

    fn get_local_obj(&self) -> LocalObj {
        self.obj.get_local_obj()
    }

    fn get_name(&self) -> String {
        self.obj.get_name()
    }

    fn get_hash_path(&self) -> String {
        self.obj.get_hash_path()
    }
}

impl Obj {
    //fn new(name: &str) -> Self {
    //    Obj { 
    //        name: String::from(name),
    //        ref_path: None, 
    //        obj_path: PathBuf::new(),
    //        file_path: PathBuf::new(),
    //        obj_type: ObjType::DEFAULT,
    //    }
    //}

    pub fn from_path(path: PathBuf) -> Self {
        let mut hasher = Sha1::new();
        hasher.input_str(path.to_str().unwrap());
        let hash = hasher.result_str();

        let (dir, res) = hash.split_at(2);
        let mut obj_path = path::objects();
        obj_path.push(dir);
        obj_path.push(res);

        let root = path::repo_root();
        let abs_path = root.join(path.clone());
        Obj { 
            name: match abs_path.file_name() {
                None => String::new(),
                Some(name) => name.to_str().unwrap().to_owned()
            },
            ref_path: None, 
            obj_path,
            obj_type: match path.exists() {
                true => match path.is_dir() {
                    true => ObjType::TREE,
                    false => ObjType::BLOB
                },
                false => ObjType::DEFAULT
            },
            file_path: abs_path,
            relative_file_path: path,
            hash_path: hash,
        }
    }

    pub fn from_line(line: String) -> Self {
            
    }

    pub fn from_head() -> Self {
        Obj { 
            name: String::new(),
            ref_path: None, 
            obj_path: head::path(),
            obj_type: ObjType::TREE,
            file_path: PathBuf::new(),
            relative_file_path: PathBuf::new(),
            hash_path: String::new(),
        }
    }



    //fn spawn(&self, kind: &str) -> Box<dyn FnOnce(Self)> {
    //    match kind {
    //        "Blob" => Box::new(|_| Blob::new(self.clone())),
    //        "Tree" => Box::new(|_| Tree::new(self.clone())),
    //        _ => unreachable!(),
    //    }
    //}
}

